/**********************************************************************
Copyright (c) 2009 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.json;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.PersistenceConfiguration;
import org.datanucleus.exceptions.NucleusDataStoreException;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues2;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.Type;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.json.fieldmanager.FetchFieldManager;
import org.datanucleus.store.json.fieldmanager.InsertFieldManager;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class AmazonS3PersistenceHandler extends JsonPersistenceHandler
{
    AmazonS3PersistenceHandler(StoreManager storeMgr)
    {
        super(storeMgr);
    }
    
    public void insertObject(ObjectProvider sm)
    {
        // Check if read-only so update not permitted
        storeMgr.assertReadOnlyForUpdateOfObject(sm);

        Properties options = new Properties();
        options.put(ConnectionFactoryImpl.STORE_JSON_URL, "/");
        ManagedConnection mconn = storeMgr.getConnection(sm.getExecutionContext(), options);
        URLConnection conn = (URLConnection) mconn.getConnection();
        createBucket(conn, getHeaderForBucket());

        int[] fieldNumbers = sm.getClassMetaData().getAllMemberPositions();
        options.put(ConnectionFactoryImpl.STORE_JSON_URL, getURLPath(sm));
        mconn = storeMgr.getConnection(sm.getExecutionContext(), options);
        conn = (URLConnection) mconn.getConnection();
        JSONObject jsonobj = new JSONObject();

        sm.provideFields(fieldNumbers, new InsertFieldManager(sm, jsonobj));
        sm.provideFields(sm.getClassMetaData().getPKMemberPositions(), new InsertFieldManager(sm, jsonobj));

        write("PUT",conn.getURL().getPath(),conn,jsonobj, getHeaders("PUT",options));
    }

    protected void createBucket(URLConnection conn, Map headers)
    {
        try
        {
            if (JSONLogger.LOGGER.isDebugEnabled())
            {
                JSONLogger.LOGGER.debug("Creating bucket. ");
            }
            HttpURLConnection http = (HttpURLConnection)conn;
            Iterator iterator = headers.keySet().iterator();
            while(iterator.hasNext())
            {
                String key = (String) iterator.next();
                String value = (String) headers.get(key);
                http.setRequestProperty(key, value);
            }
            http.setRequestProperty("Content-Length", "0");
            http.setDoOutput(true);
            http.setRequestMethod("PUT");
            http.setReadTimeout(10000);
            http.setConnectTimeout(10000);
            http.connect();
            int code = http.getResponseCode();
            
            if(code>=400)
            {
                StringBuffer sb = new StringBuffer();
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    byte[] buffer = new byte[1024];
                    int r;
                    while ((r = http.getErrorStream().read(buffer)) != -1) {
                        baos.write(buffer, 0, r);
                    }
                    
                    sb.append(new String(baos.toByteArray()));        
                http.getErrorStream().close();
                throw new NucleusDataStoreException("HTTP Error code: "+code+" "+http.getResponseMessage()+" error: "+sb.toString());
            }
            if(code>=300)
            {
                throw new NucleusDataStoreException("Redirect not supported. HTTP Error code: "+code+" "+http.getResponseMessage());
            }
        }
        catch (IOException e)
        {
            throw new NucleusDataStoreException(e.getMessage(),e);
        }
    }
    
    protected Map getHeaders(String httpVerb, Map options)
    {
        Map headers = new HashMap();
        headers.put("Date", AmazonS3Utils.getHTTPDate());
        String contentMD5 = "";
        String contentType = "";
        String urlStr = storeMgr.getConnectionURL();
        String awsKey = storeMgr.getConnectionUserName();
        String awsSecretKey = storeMgr.getConnectionPassword();

        try
        {
            urlStr = urlStr.substring(urlStr.indexOf(storeMgr.getStoreManagerKey()+":")+storeMgr.getStoreManagerKey().length()+1);
            headers.put("Host", getBucket()+"."+new URL(urlStr).getHost());
            String path = "/"+getBucket();
            if(options.containsKey(ConnectionFactoryImpl.STORE_JSON_URL))
            {
                if(!options.get(ConnectionFactoryImpl.STORE_JSON_URL).toString().startsWith("/"))
                {
                    path += "/";
                }
                if(options.get(ConnectionFactoryImpl.STORE_JSON_URL).toString().indexOf("?")>-1)
                {
                    path += options.get(ConnectionFactoryImpl.STORE_JSON_URL).toString().substring(0,options.get(ConnectionFactoryImpl.STORE_JSON_URL).toString().indexOf("?"));
                }
                else
                {
                    path += options.get(ConnectionFactoryImpl.STORE_JSON_URL).toString();
                }
            }
            else
            {
                path += new URL(urlStr).getPath();
            }
            String stringToSign = httpVerb + "\n" + contentMD5 + "\n" + contentType + "\n" + headers.get("Date") + "\n" + path;
            headers.put("Authorization", "AWS "+awsKey+":"+AmazonS3Utils.hmac(awsSecretKey,stringToSign));
        }
        catch (MalformedURLException e)
        {
            throw new NucleusException(e.getMessage(), e);
        }
        return headers;
    }
    
    protected Map getHeaderForBucket()
    {
        Map headers = new HashMap();
        headers.put("Date", AmazonS3Utils.getHTTPDate());
        String contentMD5 = "";
        String contentType = "";
        String httpVerb = "PUT";
        String urlStr = storeMgr.getConnectionURL();
        String awsKey = storeMgr.getConnectionUserName();
        String awsSecretKey = storeMgr.getConnectionPassword();
        try
        {
            urlStr = urlStr.substring(urlStr.indexOf(storeMgr.getStoreManagerKey()+":")+storeMgr.getStoreManagerKey().length()+1);
            headers.put("Host", getBucket()+"."+new URL(urlStr).getHost());
            String stringToSign = httpVerb + "\n" + contentMD5 + "\n" + contentType + "\n" + headers.get("Date") + "\n" + "/"+getBucket() +"/";
            headers.put("Authorization", "AWS "+awsKey+":"+AmazonS3Utils.hmac(awsSecretKey,stringToSign));
        }
        catch (MalformedURLException e)
        {
            e.printStackTrace();
        }
        return headers;
    }    

    /**
     * Convenience method to get all objects of the candidate type from the specified connection.
     * @param om ObjectManager
     * @param mconn Managed Connection
     * @param candidateClass Candidate
     * @param ignoreCache Whether to ignore the cache
     * @return List of objects of the candidate type
     */
    public List getObjectsOfCandidateType(ExecutionContext om, ManagedConnection mconn, 
            Class candidateClass, boolean ignoreCache, Map options)
    {
        List results = new ArrayList();

        try
        {
            URLConnection conn = (URLConnection) mconn.getConnection();
            ClassLoaderResolver clr = om.getClassLoaderResolver();
            final AbstractClassMetaData cmd = om.getMetaDataManager().getMetaDataForClass(candidateClass, clr);

            JSONArray jsonarray;
            try
            {                
                HttpURLConnection http = (HttpURLConnection) conn;
                Map headers = getHeaders("GET", options);
                Iterator iterator = headers.keySet().iterator();
                while(iterator.hasNext())
                {
                    String key = (String) iterator.next();
                    String value = (String) headers.get(key);
                    http.setRequestProperty(key, value);
                }
                http.setDoInput(true);
                http.setRequestMethod("GET");
                http.setReadTimeout(10000);
                http.setConnectTimeout(10000);
                http.connect();
                int code = http.getResponseCode();
                if (code == 404)
                {
                    return Collections.EMPTY_LIST;
                }
                handleHTTPErrorCode(http);
                StringBuffer sb = new StringBuffer();
                if (http.getContentLength() > 0)
                {
                    for (int i = 0; i < http.getContentLength(); i++)
                    {
                        sb.append((char) http.getInputStream().read());
                    }
                }
                else
                {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    byte[] buffer = new byte[1024];
                    int r;
                    while ((r = http.getInputStream().read(buffer)) != -1)
                    {
                        baos.write(buffer, 0, r);
                    }
                    sb.append(new String(baos.toByteArray()));
                }
                http.getInputStream().close();
                String contentType = http.getHeaderField("content-type");
                if(contentType!=null && (contentType.equalsIgnoreCase("application/xml") || contentType.equalsIgnoreCase("text/xml")) )
                {
                    Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new ByteArrayInputStream(sb.toString().getBytes()));
                    NodeList list = doc.getElementsByTagName("Contents");
                    Set set = new HashSet();
                    for (int i=0; i<list.getLength(); i++)
                    {
                        Element el = (Element) list.item(i);
                        JSONObject object = new JSONObject();
                        String keyText = el.getElementsByTagName("Key").item(0).getTextContent();
                        if(keyText.indexOf("/") < 1)
                        {
                            //log ignoring this element
                        }
                        else
                        {
                            String key = keyText.substring(keyText.indexOf("/",1)+1);
                            if(key.length()<1)
                            {
                                //log ignoring this element
                            }
                            else
                            {
                                String className = keyText.substring(0,keyText.indexOf("/"));
                                object.put("class", className);
                                object.put(cmd.getPrimaryKeyMemberNames()[0], key);
                                set.add(object);
                            }
                        }
                    }
                    jsonarray = new JSONArray(set);
                }
                else
                {
                    jsonarray = new JSONArray(sb.toString());
                }
            }
            catch (IOException e)
            {
                throw new NucleusDataStoreException(e.getMessage(), e);
            }
            catch (JSONException e)
            {
                throw new NucleusDataStoreException(e.getMessage(), e);
            }
            catch (SAXException e)
            {
                throw new NucleusDataStoreException(e.getMessage(), e);
            }
            catch (ParserConfigurationException e)
            {
                throw new NucleusDataStoreException(e.getMessage(), e);
            }

            for (int i = 0; i < jsonarray.length(); i++)
            {
                final JSONObject json = jsonarray.getJSONObject(i);

                results.add(om.findObjectUsingAID(new Type(clr.classForName(cmd.getFullClassName())), new FieldValues2()
                {
                    // StateManager calls the fetchFields method
                    public void fetchFields(ObjectProvider sm)
                    {
                        //we only replace the primary key, since it was the only field loaded
                        sm.replaceFields(cmd.getPKMemberPositions(), new FetchFieldManager(sm, json));
                    }

                    public void fetchNonLoadedFields(ObjectProvider sm)
                    {
                        //we only replace the primary key, since it was the only field loaded
                        sm.replaceNonLoadedFields(cmd.getPKMemberPositions(), new FetchFieldManager(sm, json));
                    }

                    public FetchPlan getFetchPlanForLoading()
                    {
                        return null;
                    }
                }, ignoreCache, true));
            }
        }
        catch (JSONException je)
        {
            je.printStackTrace();
        }

        return results;
    }

    /**
     * URL path for querying in amazon. it lists all entries for the bucket+a prefix
     */
    public String getURLPathForQuery(AbstractClassMetaData acmd)
    {
        String url = acmd.getValueForExtension("url");
        if (url==null)
        {
            url = acmd.getFullClassName();
        }        
        url = "?prefix="+url;
        return url;
    }
    
    private String getBucket()
    {
        PersistenceConfiguration conf = storeMgr.getOMFContext().getPersistenceConfiguration();
        return conf.getStringProperty("datanucleus.json.amazons3.bucket");
    }
}
